
/*
 * Copyright (c) 2011 LAAS/CNRS                       --  Oct 2011
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * Description: low-pass, Kalman and other Bayesian filters for perception
 *  
 *  
 *  
 *  Creation date : Jan 19 22:35:29 2011 
 *  Author: HE Wuwei
 *  LAAS-CNRS, Toulouse FRANCE
 */

/*
 * Matrix and basic numerical support for the Unscented Kalman Filter
 * (Basic, Scaled and Spherical Simplex)
 * 
 * The matrix part is copied (and modified) from the Bayes++ library by 
 * Copyright (c) 2002 Michael Stevens
 * And modified by HE Wuwei wuwei.he@laas.fr
 * Nov. 2011
 * 
 */


/*
 * Matrix support functions for filter classes
 * Relies on Bayesian_filter::Bayes_filter for exception
 * thrown by internal matrix checks
 */
#include "sensFlt.hpp"		// Exceptions required
#include "matSup.hpp"
#include <cassert>
#ifndef NDEBUG
#include "boost/numeric/ublas/io.hpp"
#endif



namespace {

template <class scalar>
inline scalar sqr(scalar x)
// Square 
{
	return x*x;
}
};//namespace


/* Filter Matrix Namespace */
namespace sensFusion_matrix
{


#ifndef NDEBUG
void assert_isPSD (const SymMatrix &M)
/*
 * Assert a Matrix is Positive Semi Definate via the alogrithm in isPSD
 *  Requires 'cerr' and 'assert' to about execution
 */
{
	bool bPSD = isPSD(M);
	if (!bPSD) {
		// Display Non PSD
		std::cerr.flags(std::ios::scientific); std::cerr.precision(17);
		std::cerr << M;
	}
	assert(bPSD);
}
#endif


bool isPSD (const SymMatrix &M)
/*
 * Check a Matrix is both Symertic and Positive Semi Definate
 *  Creates a temporary copy
 *  
 * Numerics of Algorithm:
 *  Use UdUfactor and checks reciprocal condition number >=0
 *  Algorithms using UdUfactor will will therefore succeed if isPSD(M) is true
 *  Do NOT assume because isPSD is true that M will appear to be PSD to other numerical algorithms
 * Return:
 *  true iff M isSymetric and is PSD by the above algorithm
 */
{
	RowMatrix UD(M.size1(),M.size1());

	RowMatrix::value_type rcond = UdUfactor(UD, M);
	return rcond >= 0.;
}


bool isSymmetric (const Matrix &M)
/*
 * Check a Symmetric Matrix really is Square and Symmetric
 * The later may be implied by the SymMatrix type
 * Implictly also checks matrix is without IEC 559 NaN values as they are always !=
 * Return:
 *  true iff M is Square and Symmetric and without NaN 
 */
{
	// Check Square
	if (M.size1() != M.size2() ) {
		return false;
	}

	// Check equality of upper and lower
	bool bSym = true;
	std::size_t size = M.size1();
	for (std::size_t r = 0; r < size; ++r) {
		for (std::size_t c = 0; c <= r; ++c) {
			if( M(r,c) != M(c,r) ) {
				bSym = false;
			}
		}
	}
	return bSym;
}


void forceSymmetric (Matrix &M, bool bUpperToLower)
/*
 * Force Matrix Symmetry
 *	Normally Copies ower triangle to upper or
 *   Upper to Lower triangle is specifed
 */
{
	// Check Square

  /*
   *
   *  Big Problem here: when use error(), compile in the library but not 
   *  niut-genom which uses the library. 
   *
   *  message: 
   *   /home/whe/openrobots/lib/libsensFusion-libs.so: undefined reference to `sensFusion_filters::Bayes_base::error(sensFusion_filters::Logic_exception const&)'
   *
   *
   */

	// if (M.size1() != M.size2() ) {
	//   using namespace sensFusion_filters;
	//   Bayes_base::error (Logic_exception ("Matrix is not square"));
	// }

	std::size_t size = M.size1();

	if (bUpperToLower)
	{
		// Copy Lower to Upper
		for (std::size_t r = 1; r < size; ++r) {
			for (std::size_t c = 0; c < r; ++c) {
				M(c,r) = M(r,c);
			}
		}
	}
	else
	{
		// Copy Upper to Lower
		for (std::size_t r = 1; r < size; ++r) {
			for (std::size_t c = 0; c < r; ++c) {
				M(r,c) = M(c,r);
			}
		}
	}
}



}//namespace
